#include <linux/in.h>
#include <linux/net.h>
#include <linux/types.h>
#include <linux/time.h>
#include <asm/errno.h>
#include <asm/fcntl.h>
#include <asm/unistd.h>

static int errno = 0;

#define __NR_CLOSE       __NR_close
static __inline__ _syscall1(int,CLOSE,int,a1)
#define __NR_DUP2        __NR_dup2
static __inline__ _syscall2(int,DUP2,int,a1,int,a2)
#define __NR_EXIT        __NR_exit
static __inline__ _syscall1(int,EXIT,int,a1)
#define __NR_FCNTL       __NR_fcntl
static __inline__ _syscall3(long,FCNTL,int,a1,int,a2,unsigned long,a3)
#define __NR_FORK        __NR_fork
static __inline__ _syscall0(int,FORK)
#define __NR_SELECT      __NR__newselect
static __inline__ _syscall5(int,SELECT,int,a1,fd_set *,a2,fd_set *,a3,fd_set *,a4,struct timeval *,a5)
#define __NR_NANOSLEEP   __NR_nanosleep
static __inline__ _syscall2(int,NANOSLEEP,struct timespec *,a1,struct timespec *,a2)
#define __NR_WAIT4       __NR_wait4
static __inline__ _syscall4(int,WAIT4,int,a1,int *,a2,int,a3,void *,a4)
#define __NR_WRITE       __NR_write
static __inline__ _syscall3(int,WRITE,int,a1,const char *,a2,off_t,a3)
#define __NR_SOCKETCALL __NR_socketcall
static __inline__ _syscall2(int,SOCKETCALL,int,a1,unsigned long *,args)

static unsigned long sargs[3] = { PF_INET, SOCK_STREAM, IPPROTO_TCP };
static unsigned long cargs[3] = { 1, 0, sizeof(struct sockaddr_in) };
static unsigned long gargs[3] = { 0, 0, 0 };

#define SOCKET()      SOCKETCALL(SYS_SOCKET, &sargs[0])
#define CONNECT()     SOCKETCALL(SYS_CONNECT, &cargs[0])
#define GETPEERNAME() SOCKETCALL(SYS_GETPEERNAME, &gargs[0])

static unsigned char buf[5] = "0000\n";
#define hhex(X,Y) ((Y=((X)>>4))>9?Y+'a'-10:Y+'0')
#define lhex(X,Y) ((Y=((X)&15))>9?Y+'a'-10:Y+'0')

/* splitting these off as functions makes the executable smaller since
 * gcc is smarter about optimizing them, even though they are only called
 * once (using gcc 2.7.2.3)
 */
static __inline__ void note_net(unsigned int i) {
  int j;

  buf[0] = hhex(i, j);
  buf[1] = lhex(i, j);
}

static __inline__ void quick_write(unsigned int i) {
  int j;

  buf[2] = hhex(i, j);
  buf[3] = lhex(i, j);
  WRITE(0, (char *)buf, 5);
}

extern void _start() {
/* network byte order */
#define PORT     0x1900
#define CLASSB   0x00005f82
  static struct timespec ts;
  static unsigned int i;
  static int p;
  static struct sockaddr_in saddr;
  static fd_set fds, rf, wf;
  static struct timeval tm;

  gargs[1] = cargs[1] = (unsigned long)&saddr;
  gargs[2] = (unsigned long)&ts.tv_sec;
  tm.tv_sec = 30;
  tm.tv_usec = 0;
  FD_ZERO(&fds);
  for(i=1;i<=254;i++)
    FD_SET(i, &fds);
  for(i=0;i<=255;i++)
    if(i != 1)
      CLOSE(i);
  DUP2(1, 0);
  CLOSE(1);
  saddr.sin_family = AF_INET;
  saddr.sin_port = PORT;
  for(i=0; i<=255; i++) {
    saddr.sin_addr.s_addr = CLASSB | (i<<16);
    if((p = FORK())) {
      if(p == -1) {
        goto ERR_EXIT;
      } else {
        while(WAIT4(-1, &p, WNOHANG, NULL) > 0);
        ts.tv_sec = 1;
        ts.tv_nsec = 0;
        NANOSLEEP(&ts,&ts);
      }
    } else {
      break;
    }
  }
  if(i == 256) {
    while(WAIT4(-1, &p, 0, NULL) != -1 && errno != EINTR);
    if(errno != ECHILD)
      goto ERR_EXIT;
  } else {
    note_net(i);
    for(i=1; i<=254; i++, cargs[0]++) {
      saddr.sin_addr.s_addr |= i<<24;
      if(SOCKET() != i ||
         FCNTL(i, F_SETFL, O_RDWR|O_NONBLOCK) == -1 ||
         CONNECT() != -1 ||
         errno != EINPROGRESS)
        goto ERR_EXIT;
      saddr.sin_addr.s_addr &= ~0xff000000L;
    }
    rf = wf = fds;
    while((p=SELECT(255, &rf, &wf, NULL, &tm)) > 0) {
      for(i=1; p && i <= 254; i++)
        if(FD_ISSET(i, &rf) || FD_ISSET(i, &wf)) {
          p--;
          ts.tv_sec = sizeof(saddr);
          FD_CLR(i, &fds);
          gargs[0] = i;
          if(GETPEERNAME() == 0)
            quick_write(i);
          CLOSE(i);
        }
      rf = wf = fds;
    }
    if(p == 0)
      goto EXIT;
  }
ERR_EXIT:
  WRITE(0, "XXXX\n", 5);
EXIT:
  EXIT(0);
}
